/* 
   Demonstrate usage of double linked lists routines LL for float numbers 
   
   author: G. Matas (g.matas@ee.surrey.ac.uk)

*/
/*-----------------------------------------------------------------------*/
static char sccsid[]="@(#)93/06/30 g.matas@ee.surrey.ac.uk 5.1 ex2.c";


#include "LL.h"


/* function applied to individual elements */
static void * PrintDouble(void * p_elm)
	  { printf("%4.2g",*(double*)p_elm); return NULL; }


static int CompDouble (const void * p_elm1,const  void *  p_elm2)
{ return (* (double const *)p_elm1<= * (double const *)p_elm2) ? -1 : 1 ; }


int main (int argc, char ** argv)
{
  double some_num [] = { 5.6, 3.9, 4.2, 6.7};
  double * dp, num;

  t_LL   list = ConsLL();         /*                   create an empty list  */


  { /* insert all numbers in some_num  into the list  */
  int i;
  for (i=0; i < sizeof(some_num)/sizeof(double); i++)
    InsLastLL(list,some_num[i]);
  }


  /* add some other numbers to the end of the list  */
    
  num = 7.15;
  dp = InsLastLL(list,num);
  num = 8.00;
  dp = InsLastLL(list,num);

  /* sort the list  and print the results */
  SortLL(list, CompDouble);
  printf("Double numbers sorted:\n");
  ApplyLL (list, PrintDouble);
  printf("\n\n");


  /* insert some arbitrary numbers at the beggining */
  num = 1.0;	 InsFirstLL(list, num);
  num = -1.0;    InsFirstLL(list, num);

  /* sort and print again */
    SortLL(list, CompDouble);
    printf("Double numbers sorted second time:\n");
    ApplyLL (list, PrintDouble);
    printf("\n\n");
	
  /* test CopyAftLL, MoveAftLL   */
    {
    t_LL copied_list=ConsCopyLL(list);
    num = 9.99; InsFirstLL(copied_list,num);
    printf("copied list  \n");
    ApplyLL (copied_list, PrintDouble);
    printf("\n\n");

    MoveListLastLL(list,copied_list);
    printf("Two times repeated:\n");
    DestLL(copied_list);
    }  /* end of copied list scope*/

    ApplyLL (list, PrintDouble);
    printf("\n\n");


    SortLL(list, CompDouble);
    printf("Double numbers sorted second time:\n");
    ApplyLL (list, PrintDouble);
    printf("\n\n");

  /* to check the backward chaining, print out in reverse order */
   {
   double * el;

   printf("Printed in reverse order\n");
   ForeachDownLL_M(list,el)
    printf("%4.2g ",*el);
   printf("\n\n");
   }


   /* and Split now  */
   {
    t_LL list2 = MoveTailLastLL(ConsLL(),list,NthElmLL(list,6));

    printf("Elements 1-5, then the rest\n");
    ApplyLL (list, PrintDouble);
    printf("\n");
    ApplyLL (list2, PrintDouble);
    printf("\n\n");
    
    MoveListFirstLL(list,list2);
    MoveHeadFirstLL(list2,list,NthElmLL(list,3));
    printf("Elements 6- first, 1&2 cut out\n");
    ApplyLL (list, PrintDouble);
    printf("\n");
    ApplyLL (list2, PrintDouble);
    printf("\n\n");
    
    MoveHeadLastLL(list,list2,NextElmLL(LastElmLL(list2)));
    printf("should be everything...\n");
    ApplyLL (list, PrintDouble);
    printf("\n");
    
    MoveTailLastLL(list2,list,PrevElmLL(FirstElmLL(list2)));
    printf("should be everything...\n");
    ApplyLL (list, PrintDouble);
    printf("\n");
    


    DestLL(list2);   /* destroy the list before it goes out of scope */
   }
    
   DestLL(list);

   return 0;   
}
